from datetime import datetime, timedelta, timezone
import logging
import socket
import ssl
import threading

import appdirs

from autocert.acme import ACMEClient
from autocert.cache import Cache
from autocert.errors import AutocertError
from autocert.keys import PrivateKey
from autocert.manager import Manager

log = logging.getLogger(__name__)


def manage(sock, *domains, contact=None, accept_tos=False):
    # ensure args are valid
    if not accept_tos:
        raise AutocertError("CA's Terms of Service must be accepted")
    if not isinstance(sock, socket.socket):
        raise AutocertError('Socket sock must be a socket')

    # if contact email is just a string, make it a single-element list
    if type(contact) == str:
        contact = [contact]

    # use a platform-friendly directory for caching keys / certs
    cache_dir = appdirs.user_cache_dir('python-autocert', 'python-autocert')
    log.info('using cache dir: %s', cache_dir)
    cache = Cache(cache_dir)

    # ensure account pkey exists
    account_pkey_name = 'account_acme.pkey'
    if cache.exists(account_pkey_name):
        log.info('loading existing account key: %s', account_pkey_name)
        account_pkey_pem = cache.read(account_pkey_name)
        account_pkey = PrivateKey(account_pkey_pem)
    else:
        log.info('generating new account key: %s', account_pkey_name)
        account_pkey = PrivateKey()
        account_pkey_pem = account_pkey.pem
        cache.write(account_pkey_name, account_pkey_pem)

    # first domain in the list will be the subject (all domains will be SANs)
    primary_domain = domains[0]
    tls_pkey_name = primary_domain + '.pkey'
    tls_cert_name = primary_domain + '.cert'

    # read or create TLS pkey
    tls_pkey = None
    if cache.exists(tls_pkey_name):
        tls_pkey_pem = cache.read(tls_pkey_name)
        tls_pkey = PrivateKey(tls_pkey_pem)
    else:
        tls_pkey = PrivateKey()
        tls_pkey_pem = tls_pkey.pem
        cache.write(tls_pkey_name, tls_pkey_pem)

    # ensure TLS cert exists (defaults to a self-signed cert)
    if not cache.exists(tls_cert_name):
        tls_cert_pem = tls_pkey.generate_self_signed_cert(domains)
        cache.write(tls_cert_name, tls_cert_pem)

    # create ssl context w/ modern ciphers/options
    ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)

    # SSLContext.load_cert_chain wants real files :(
    tls_pkey_path = cache.path(tls_pkey_name)
    tls_cert_path = cache.path(tls_cert_name)
    ctx.load_cert_chain(tls_cert_path, tls_pkey_path)

    # init ACME client and cert manager
    client = ACMEClient(account_pkey, contact=contact, accept_tos=accept_tos)
    manager = Manager(tls_pkey, ctx, cache, domains, client)

    # hook manager into the context
    ctx._msg_callback = manager.msg_callback
    ctx.sni_callback = manager.sni_callback

    # kick off cert issuance and renewal loop in a background thread
    thread = threading.Thread(
        target=manager.issue_and_renew_forever,
        daemon=True,
    )
    thread.start()

    # wrap and return the TLS-enabled socket
    sock_tls = ctx.wrap_socket(sock, server_side=True)
    return sock_tls



from datetime import datetime, timedelta, timezone
import hashlib
import logging

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import utils as crypto_utils
from cryptography.x509.oid import ExtendedKeyUsageOID, NameOID

from autocert import utils

# OID for the ACME extension for the TLS-ALPN challenge.
# https://tools.ietf.org/html/draft-ietf-acme-tls-alpn-05#section-5.1
ID_PE_ACME_IDENTIFIER = x509.ObjectIdentifier('1.3.6.1.5.5.7.1.31')

log = logging.getLogger(__name__)


class PrivateKey:

    def __init__(self, pem=None):
        if pem is not None:
            self.key = serialization.load_pem_private_key(pem, password=None)
            self.pem = pem
        else:
            self.key = ec.generate_private_key(curve=ec.SECP256R1())
            self.pem = self.key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )

        self.public_key = PublicKey(self.key.public_key())

    def sign(self, data):
        # https://community.letsencrypt.org/t/parse-error-reading-jws/137654/13
        signature = self.key.sign(data, ec.ECDSA(hashes.SHA256()))
        r, s = crypto_utils.decode_dss_signature(signature)
        r = utils.int_to_bytes(r)
        s = utils.int_to_bytes(s)
        signature = r + s
        return signature

    def generate_csr(self, domains):
        # https://cryptography.io/en/latest/x509/reference.html#x-509-csr-certificate-signing-request-builder-object
        builder = x509.CertificateSigningRequestBuilder()
        builder = builder.subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
        ]))
        builder = builder.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        builder = builder.add_extension(
            x509.SubjectAlternativeName(
                [x509.DNSName(domain) for domain in domains]
            ),
            critical=True,
        )

        # sign the cert and convert to DER
        csr = builder.sign(private_key=self.key, algorithm=hashes.SHA256())
        csr = csr.public_bytes(serialization.Encoding.DER)
        return csr

    def generate_self_signed_cert(self, domains, ttl=timedelta(days=30)):
        # https://cryptography.io/en/latest/x509/reference.html#x-509-certificate-builder
        builder = x509.CertificateBuilder()
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
        ]))
        builder = builder.issuer_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
        ]))
        builder = builder.not_valid_before(datetime.now(timezone.utc))
        builder = builder.not_valid_after(datetime.now(timezone.utc) + ttl)
        builder = builder.public_key(self.key.public_key())
        builder = builder.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        builder = builder.add_extension(
            x509.SubjectAlternativeName(
                [x509.DNSName(domain) for domain in domains]
            ),
            critical=True
        )

        # sign the cert and convert to PEM
        cert = builder.sign(private_key=self.key, algorithm=hashes.SHA256())
        cert = cert.public_bytes(serialization.Encoding.PEM)
        return cert

    def generate_tls_alpn_01_cert(self, domain, keyauth, ttl=timedelta(days=30)):
        # create the ACME identifier
        acme_identifier = hashlib.sha256(keyauth).digest()
        acme_identifier = utils.bytes_to_der(acme_identifier)

        # https://cryptography.io/en/latest/x509/reference.html#x-509-certificate-builder
        builder = x509.CertificateBuilder()
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, domain),
        ]))
        builder = builder.issuer_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, domain),
        ]))
        builder = builder.not_valid_before(datetime.now(timezone.utc))
        builder = builder.not_valid_after(datetime.now(timezone.utc) + ttl)
        builder = builder.public_key(self.key.public_key())
        builder = builder.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        builder = builder.add_extension(
            x509.SubjectAlternativeName([
                x509.DNSName(domain),
            ]),
            critical=True,
        )
        builder = builder.add_extension(
            x509.ExtendedKeyUsage([
                ExtendedKeyUsageOID.SERVER_AUTH,
            ]),
            critical=True,
        )
        builder = builder.add_extension(
            # https://github.com/pyca/cryptography/issues/2747
            x509.UnrecognizedExtension(ID_PE_ACME_IDENTIFIER, acme_identifier),
            critical=True,
        )

        # sign the cert and convert to PEM
        cert = builder.sign(private_key=self.key, algorithm=hashes.SHA256())
        cert = cert.public_bytes(serialization.Encoding.PEM)
        return cert


class PublicKey:
    NIST_CURVE_NAMES = {
        'secp192r1': 'P-192',
        'secp224r1': 'P-224',
        'secp256r1': 'P-256',
        'secp384r1': 'P-384',
        'secp521r1': 'P-521',
    }

    def __init__(self, key):
        self.key = key
        self.pem = self.key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    def curve(self):
        curve = self.key.curve.name
        curve = self.NIST_CURVE_NAMES[curve]
        return curve

    @property
    def x(self):
        x = self.key.public_numbers().x
        x = utils.int_to_bytes(x)
        return x

    @property
    def y(self):
        y = self.key.public_numbers().y
        y = utils.int_to_bytes(y)
        return y



from datetime import datetime, timedelta, timezone
import logging
import ssl
import time

from cryptography import x509

from autocert.errors import AutocertError

log = logging.getLogger(__name__)


class Manager:

    def __init__(self, private_key, context, cache, domains, client):
        self.private_key = private_key
        self.context = context
        self.cache = cache
        self.domains = domains
        self.client = client

        # setup primary domain and key / cert names
        self.primary_domain = domains[0]
        self.tls_pkey_name = self.primary_domain + '.pkey'
        self.tls_cert_name = self.primary_domain + '.cert'

        # setup key / cert file paths
        self.tls_pkey_path = self.cache.path(self.tls_pkey_name)
        self.tls_cert_path = self.cache.path(self.tls_cert_name)

        # flags to track when a TLS-ALPN-01 challenge is expected / in-progress
        self.expecting_challenge = False
        self.acme_tls_challenge = False

    def issue_and_renew_forever(self):
        log.info('starting issue/renew loop for domains: %s', self.domains)
        while True:
            # read cert from cache
            cert_pem = self.cache.read(self.tls_cert_name)

            # load as x509 and check TTL
            cert = x509.load_pem_x509_certificate(cert_pem)
            expiry = cert.not_valid_after
            expiry = expiry.replace(tzinfo=timezone.utc)

            # determine how many seconds remain until TTL is within 30 days
            remaining = expiry - datetime.now(timezone.utc) - timedelta(days=30)
            remaining = remaining.total_seconds()

            # sleep til the 30 day mark
            if remaining > 0:
                log.info('cert is still valid, sleeping for: %s', remaining)
                time.sleep(remaining)

            # time to issue / renew
            log.info('time is up, renewing cert for: %s', self.domains)
            self.issue_and_renew()

    def issue_and_renew(self):
        from pprint import pprint

        # TODO: check if order['status'] isn't pending for some reason
        order = self.client.create_order(self.domains)
        pprint(order)

        auth_urls = order['authorizations']
        for auth_url in auth_urls:
            auth = self.client.get_authorization(auth_url)
            pprint(auth)

            # pull out the domain and TLS-ALPN-01 challenge
            domain = auth['identifier']['value']
            challenge = [c for c in auth['challenges'] if c['type'] == 'tls-alpn-01'][0]

            # determine the keyauth value
            token = challenge['token']
            keyauth = self.client.get_keyauth(token)

            # generate the TLS-ALPN-01 challenge chain
            acme_cert_name = domain + '.acme.cert'
            acme_cert_pem = self.private_key.generate_tls_alpn_01_cert(domain, keyauth)
            self.cache.write(acme_cert_name, acme_cert_pem)

            # get ready for challenge requests
            self.expecting_challenge = True

            # tell ACME server to verify our challenge
            challenge_url = challenge['url']
            self.client.verify_challenge(challenge_url)

            # poll til status isn't pending anymore
            # TODO: smarter backoff here
            auth = self.client.get_authorization(auth_url)
            while auth['status'] == 'pending':
                time.sleep(1)
                auth = self.client.get_authorization(auth_url)

            # challenge is over
            self.expecting_challenge = False

            pprint(auth)

            # at this point, we either failed or passed the challenge
            if auth['status'] != 'valid':
                raise AutocertError('failed to satisfy ACME challenge: {}'.format(auth))

        # generate CSR for new cert
        csr = self.private_key.generate_csr(self.domains)

        # finalize the order
        finalize_url = order['finalize']
        order = self.client.finalize_order(finalize_url, csr)
        pprint(order)

        # download the cert
        cert_url = order['certificate']
        cert_pem = self.client.download_certificate(cert_url)

        # replace certs in the cache
        self.cache.write(self.tls_cert_name, cert_pem)

        # update the managed SSLContext with the new cert chain
        self.context.load_cert_chain(self.tls_cert_path, self.tls_pkey_path)

    def msg_callback(self, conn, direction, version, content_type, msg_type, data):
        # early exit if not expecting a challenge
        if not self.expecting_challenge:
            return

        # else look for 'acme-tls/1' in the raw stream
        if direction == 'read' and b'acme-tls/1' in data:
            self.acme_tls_challenge = True
            log.info('!!! saw an acme-tls/1 request !!!')

    def sni_callback(self, sslsocket, sni_name, sslcontext):
        # early exit if not in a challenge
        if not self.acme_tls_challenge:
            return

        # ignore empty sni_name
        if sni_name is None:
            log.info('got an empty sni_name during a TLS-ALPN-01 challenge')
            return

        # make sure TLS-ALPN-01 challenge cert exists for this sni_name
        log.info('answering TLS-ALPN-01 challenge for: %s', sni_name)
        acme_cert_name = sni_name + '.acme.cert'
        if not self.cache.exists(acme_cert_name):
            log.info('missing TLS-ALPN-01 challenge cert: %s', acme_cert_name)
            return

        # create an ephemeral SSLContext for the challenge response
        ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        ctx.set_alpn_protocols(['acme-tls/1'])

        # HACK: fix deadlock until updates come through
        ctx._msg_callback = self.msg_callback

        # serve up the TLS-ALPN-01 challenge cert
        acme_cert_path = self.cache.path(acme_cert_name)
        ctx.load_cert_chain(acme_cert_path, self.tls_pkey_path)

        # update the socket with the new context
        sslsocket.context = ctx

        # reset acme_tls_challenge flag
        self.acme_tls_challenge = False



import base64


def base64_rfc4648(b):
    return base64.urlsafe_b64encode(b).decode('utf8').replace('=', '')


def int_to_bytes(i):
    return i.to_bytes((i.bit_length() + 7) // 8, byteorder='big')


def bytes_to_der(b):
    if len(b) >= 128:
        raise ValueError('bytes_to_der only works for small bytes (< 128)')

    octet_string = 0x04
    header = [octet_string, len(b)]
    return bytes(header) + b



from urllib.parse import quote


def hello_world_app(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-Type', 'text/plain')]
    start_response(status, response_headers)
    return [b'Hello world from autocert!\n']


def redirect_app(environ, start_response):
    status = '301 Moved Permanently'
    response_headers = [('Location', reconstruct_https_url(environ))]
    start_response(status, response_headers)
    return [b'']


# https://www.python.org/dev/peps/pep-3333/#url-reconstruction
def reconstruct_https_url(environ):
    url = 'https://'

    if environ.get('HTTP_HOST'):
        url += environ['HTTP_HOST']
    else:
        url += environ['SERVER_NAME']

        if environ['wsgi.url_scheme'] == 'https':
            if environ['SERVER_PORT'] != '443':
               url += ':' + environ['SERVER_PORT']
        else:
            if environ['SERVER_PORT'] != '80':
               url += ':' + environ['SERVER_PORT']

    url += quote(environ.get('SCRIPT_NAME', ''))
    url += quote(environ.get('PATH_INFO', ''))
    if environ.get('QUERY_STRING'):
        url += '?' + environ['QUERY_STRING']

    return url



import json

from autocert import utils


class JWS(dict):

    def __init__(self, url, payload, nonce, jwk=None, kid=None):
        if jwk is None and kid is None:
            raise ValueError('either "jwk" or "kid" must be specified')

        protected = self.encode_protected(url, nonce, jwk=jwk, kid=kid)
        payload = self.encode_payload(payload)
        jws = {
            'protected': protected,
            'payload': payload,
        }

        super().__init__(jws)

    def sign(self, private_key):
        protected = self['protected']
        payload = self['payload']

        signature = '{}.{}'.format(protected, payload)
        signature = signature.encode()
        signature = private_key.sign(signature)
        signature = utils.base64_rfc4648(signature)

        self['signature'] = signature
        return json.dumps(self, separators=(',', ':'))

    def encode_protected(self, url, nonce, jwk=None, kid=None):
        if jwk is None and kid is None:
            raise ValueError('either "jwk" or "kid" must be specified')

        protected = {
            'alg': 'ES256',
            'nonce': nonce,
            'url': url,
        }
        # use kid if present else default to jwk
        if kid is not None:
            protected['kid'] = kid
        else:
            protected['jwk'] = jwk

        protected = json.dumps(protected, separators=(',', ':'), sort_keys=True)
        protected = protected.encode()
        protected = utils.base64_rfc4648(protected)
        return protected

    def encode_payload(self, payload):
        if payload is None:
            return ''

        payload = json.dumps(payload, separators=(',', ':'), sort_keys=True)
        payload = payload.encode()
        payload = utils.base64_rfc4648(payload)
        return payload 



import os


class Cache:

    def __init__(self, cache_dir):
        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def path(self, name):
        return os.path.join(self.cache_dir, name)

    def exists(self, name):
        path = os.path.join(self.cache_dir, name)
        return os.path.exists(path)

    def read(self, name):
        path = os.path.join(self.cache_dir, name)
        with open(path, 'rb') as f:
            data = f.read()
        return data

    def write(self, name, data):
        path = os.path.join(self.cache_dir, name)
        with open(path, 'wb') as f:
            f.write(data)



from setuptools import setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='autocert',
    version='0.0.4',
    author='Andrew Dailey',
    description='Automatic TLS cert issuance and renewal for Python web apps',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/theandrew168/autocert',
    packages=['autocert'],
    install_requires=[
        'appdirs',
        'cryptography',
        'requests',
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.0',
)



from datetime import datetime, timedelta, timezone
import hashlib
import logging
import os

import appdirs
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
import requests

from autocert import utils
from autocert.errors import AutocertError
from autocert.jwk import JWK
from autocert.jws import JWS

#LETS_ENCRYPT_ACME_URL = 'https://acme-v02.api.letsencrypt.org/directory'
LETS_ENCRYPT_ACME_URL = 'https://acme-staging-v02.api.letsencrypt.org/directory'

log = logging.getLogger(__name__)


class ACMEClient:

    def __init__(self, private_key, contact=None, accept_tos=False, directory_url=LETS_ENCRYPT_ACME_URL):
        if not accept_tos:
            raise AutocertError("CA's Terms of Service must be accepted")

        # initial ACME credentials
        self.private_key = private_key
        self.jwk = JWK(private_key.public_key)
        self.kid = None

        # other ACME details
        self.contact = contact
        self.accept_tos = accept_tos
        self.directory_url = directory_url

        # grab current directory and initial nonce
        self.directory = requests.get(self.directory_url).json()
        self.nonce = requests.head(self.directory['newNonce']).headers['Replay-Nonce']

        # load existing ACME account or create a new one
        acct = self._create_or_read_account()
        self.kid = acct.headers['Location']
        log.info('initialized account kid: %s', self.kid)

    def create_order(self, domains):
        log.info('creating order for domains: %s', domains)
        url = self.directory['newOrder']
        payload = {
            'identifiers': [
                {'type': 'dns', 'value': domain} for domain in domains
            ],
        }
        resp = self._cmd(url, payload)
        return resp.json()

    def get_authorization(self, auth_url):
        log.info('getting authorization: %s', auth_url)
        resp = self._cmd(auth_url, None)
        return resp.json()

    def verify_challenge(self, challenge_url):
        log.info('verifying challenge: %s', challenge_url)
        resp = self._cmd(challenge_url, {})
        return resp.json()

    def get_keyauth(self, token):
        thumbprint = self.jwk.thumbprint()
        keyauth = '{}.{}'.format(token, thumbprint)
        keyauth = keyauth.encode()
        return keyauth

    def finalize_order(self, finalize_url, csr):
        payload = {
            'csr': utils.base64_rfc4648(csr),
        }
        resp = self._cmd(finalize_url, payload)
        return resp.json()

    def download_certificate(self, cert_url):
        resp = self._cmd(cert_url, None)
        return resp.content

    def _create_or_read_account(self):
        url = self.directory['newAccount']
        payload = {
            'termsOfServiceAgreed': self.accept_tos,
        }

        # apply contact emails if present
        contact = self.contact
        if contact is not None:
            log.info('attaching contact info to account: %s', contact)

            # add mailto prefix to each email if not already present
            emails = []
            for email in contact:
                if email.startswith('mailto:'):
                    emails.append(email)
                else:
                    emails.append('mailto:' + email)

            # add to payload
            payload['contact'] = emails

        return self._cmd(url, payload)

    def _cmd(self, url, payload):
        headers = {
            'Content-Type': 'application/jose+json',
        }
        jws = JWS(url, payload, self.nonce, jwk=self.jwk, kid=self.kid)
        jws = jws.sign(self.private_key)

        # post message to the ACME server
        resp = requests.post(url, headers=headers, data=jws)
        if resp.status_code not in [200, 201, 204]:
            # TODO: if bad nonce, get another and retry
            log.error('jws: %s', jws)
            log.error('url: %s', url)
            log.error('pay: %s', payload)
            log.error('nonce: %s', self.nonce)
            raise AutocertError('ACME error: {}'.format(resp.json()))

        # update nonce
        self.nonce = resp.headers['Replay-Nonce']

        return resp



import hashlib
import json

from autocert import utils


class JWK(dict):

    def __init__(self, public_key):
        jwk = {
            'kty': 'EC',
            'crv': public_key.curve,
            'x': utils.base64_rfc4648(public_key.x),
            'y': utils.base64_rfc4648(public_key.y),
        }
        super().__init__(jwk)

    def thumbprint(self):
        thumbprint = json.dumps(self, separators=(',', ':'), sort_keys=True)
        thumbprint = thumbprint.encode()
        thumbprint = hashlib.sha256(thumbprint).digest()
        thumbprint = utils.base64_rfc4648(thumbprint)
        return thumbprint
